#include "HelMod.h"
#include "loginterp"
#include "xstream"
#include <algorithm>
#include <filesystem>
#include <limits>
#include <list>
#include <map>
#include <sstream>
#include <stdexcept>
#include <valarray>

// labeling convention:
// i(nner), o(uter): helmode matrix grid
// t(oa)  , l(is)  : user interface grid
// e(kn)  , r(ig)  : energy quantity

using HelMod::Base;
using HelMod::Isotope;
using HelMod::Modulator;

static std::map<std::string, std::map<long, std::string>> _data_ref;

std::string HelMod::IsotopeFile(int Z, int A, const std::string& data_folder) {
    long za;
    int* pza = (int*)&za;
    auto base_path = std::filesystem::weakly_canonical(
        data_folder.empty() ? _HELMOD_DATADIR_STR : data_folder);
    auto base_str = base_path.string();
    if (!std::filesystem::is_directory(base_path)) {
        throw std::runtime_error("invalid helmod data folder '" + base_str +
                                 "'");
    }
    auto entry = _data_ref.find(base_str);
    if (entry == _data_ref.end()) {
	// create list of contents
	entry = _data_ref.insert({base_str, {}}).first;
	for (const auto& p: std::filesystem::directory_iterator(base_path)) {
	    const std::string filename = p.path().filename().string();
	    // if (!filename.end_with(".xdat")) // c++20
	    if (filename.size() < 5 || ".xdat" != filename.substr(filename.size() - 5)) {
		continue;
	    }
	    std::stringstream ss(filename);
	    char _;
	    if (ss >> pza[0] >> _ >> pza[1]) {
		entry->second.insert({za, filename});
	    }
	}
    }
    pza[0] = Z;
    pza[1] = A;
    auto rec = entry->second.find(za);
    return rec == entry->second.end() ? "" : (base_path / rec->second).string();
}

void Base::_calculate_beta(void) {
    _beta_inner = std::sqrt(1. - 1. / std::pow(1. + _ekin_inner / T0, 2));
    _beta_outer = std::sqrt(1. - 1. / std::pow(1. + _ekin_outer / T0, 2));
}

std::valarray<double>
Base::SparseMatrix::apply(const std::valarray<double>& flux_outer_rig) const {
    const size_t n_inner = iout_starts.size();
    // if (flux_outer_rig.size() != n_inner) {
    //     throw std::logic_error("wrong size of `flux_outer', "
    //                            "should interpolate onto `_ekin_outer' first");
    // }
    std::valarray<double> flux_inner_rig(0., n_inner);
    for (size_t i = 0; i < n_inner; ++i) {
        const auto& row     = mat[i];
        const double istart = iout_starts[i];
        flux_inner_rig[i] =
            (row * flux_outer_rig[std::slice(istart, row.size(), 1)]).sum();
    }
    return flux_inner_rig;
}

Isotope::Isotope(const std::string& filename) {
    yuc::ixstream ifx(filename);
    if (!ifx) {
        throw std::runtime_error("unable to read binary data '" + filename +
                                 "'");
    }

    std::string isoname2;
    ifx >> isotope_name >> isoname2;
    isotope_name += isoname2;

    ifx >> *(reinterpret_cast<long*>(&isotope_number));

    ifx >> T0 >> _ekin_outer >> _ekin_inner;
    _calculate_beta();

    tstart = std::numeric_limits<size_t>::max();
    tend   = std::numeric_limits<size_t>::min();
    size_t time;
    while (ifx >> time) {
        tstart = std::min(tstart, time);
        tend   = std::max(tend, time);

        _time_slices.push_back({time, {}});
        auto& mat = _time_slices.back().second;
        ifx >> mat.iout_starts >> mat.mat;
    }
}

Modulator Isotope::ExpWindow(size_t time_start, size_t time_end) const {
    return {*this, time_start, time_end};
}

Modulator::Modulator(const Isotope& iso, size_t time_start, size_t time_end) {

    if (time_start > iso.tend || time_end < iso.tstart) {
        throw std::runtime_error(
            "modulation time window out of range. avaliable: [" +          //
            std::to_string(tstart) + ", " + std::to_string(tend) +         //
            "],  quested: ["                                               //
            + std::to_string(time_start) + ", " + std::to_string(time_end) //
            + "]");
    }

    isotope_name   = iso.isotope_name;
    isotope_number = iso.isotope_number;
    T0             = iso.T0;

    _ekin_inner = iso._ekin_inner;
    _ekin_outer = iso._ekin_outer;
    _beta_inner = iso._beta_inner;
    _beta_outer = iso._beta_outer;

    tstart = std::numeric_limits<size_t>::max();
    tend   = std::numeric_limits<size_t>::min();

    const size_t n_inner = _ekin_inner.size();
    const size_t n_outer = _ekin_outer.size();
    const size_t n_slice = iso._time_slices.size();

    _sparse_mat.iout_starts.resize(n_inner, n_outer); // (size, value)
    _sparse_mat.mat.resize(n_inner);

    auto& istarts = _sparse_mat.iout_starts;
    std::valarray<size_t> iends((size_t)0, n_inner); // (value, count)
    std::vector<size_t> time_ids;

    for (size_t islice = 0; islice < n_slice; ++islice) {
        const auto& time_slice = iso._time_slices[islice];
        const auto& time       = time_slice.first;
        const auto& slice      = time_slice.second;

        if (time < time_start || time > time_end) {
            continue;
        }

        time_ids.push_back(islice);

        tstart = std::min(tstart, time);
        tend   = std::max(tend, time);

        for (size_t j = 0; j < n_inner; ++j) {
            istarts[j] = std::min(istarts[j], slice.iout_starts[j]);
            iends[j] =
                std::max(iends[j], slice.iout_starts[j] + slice.mat[j].size());
        }
    }

    for (size_t j = 0; j < n_inner; ++j) {
        _sparse_mat.mat[j].resize(iends[j] - istarts[j]);
    }

    const size_t n_slice_in_window = time_ids.size();
    for (size_t i = 0; i < n_slice_in_window; ++i) {
        const auto& slice = iso._time_slices[time_ids[i]].second;
        std::valarray<size_t> relative_istarts = slice.iout_starts - istarts;
        for (size_t j = 0; j < n_inner; ++j) {
            const auto& row = slice.mat[j];
            _sparse_mat
                .mat[j][std::slice(relative_istarts[j], row.size(), 1)] += row;
        }
    }

    for (size_t j = 0; j < n_inner; ++j) {
        _sparse_mat.mat[j] /= n_slice_in_window;
    }

    return;
}

std::valarray<double>
Modulator::modulate(const std::valarray<double>& ekin_lis,
                    const std::valarray<double>& flux_lis,
                    const std::valarray<double>& ekin_toa) const {
    // always assume ekin_lis to envelop ekin_toa and ekin_inner

    // if all energy is high enough, dont no modulation need
    const double max_mod_e = _ekin_inner.max();
    if (ekin_lis.min() > max_mod_e) {
        return ekin_toa.size()
                   ? yuc::log_interpolator(ekin_lis, flux_lis)(ekin_toa)
                   : flux_lis;
    }

    std::valarray<double> used_ekin_toa = ekin_toa.size() ? ekin_toa : ekin_lis;

    yuc::log_interpolator intf_l_e(ekin_lis, flux_lis);
    // according to HelMod python module
    std::valarray<double> flux_o_r = intf_l_e(_ekin_outer) / _beta_outer;
    std::valarray<double> flux_i_e = _sparse_mat.apply(flux_o_r) * _beta_inner;
    yuc::log_interpolator intf_i_e(_ekin_inner, flux_i_e);

    if (used_ekin_toa.max() <= max_mod_e) {
        return intf_i_e(used_ekin_toa);
    } else {
        for (auto& e : used_ekin_toa) {
            e = (e > max_mod_e ? intf_l_e : intf_i_e)(e);
        }
        return used_ekin_toa;
    }
}
